"""A module defining all point time series extractor classes"""

import numpy as np
from netCDF4 import Dataset
from abc import ABC, abstractmethod
import datetime as dt
from tfv.misc import Expression
from tfv.geometry import Mesh


__version__ = "0.0.2"
__author__ = "toby.devlin@bmtglobal.com, jonah.chorley@bmtglobal.com"
__aus_date__ = "%d/%m/%Y %H:%M:%S"


class TimeSeries(ABC):
    """A base class which defines the API for all point time series based model result data extraction"""

    def __init__(self, file):
        """Initializes TimeSeries extractor object with model result file"""

        # Store file path string as attribute
        self.file = file
        self.sites = None

        # Prepare static TimeSeries attributes
        self.__prepare_file_handle__()
        self.__prepare_time_vector__()
        self.__prepare_locations__()

    @abstractmethod
    def get_raw_data(self, variable, site):
        """
        Query to extract raw time series data at a given location.

        Arguments
        ---------
        :param variable -- string | Name of time series data set to be extracted.
        :param site -- string | Location at which to extract the time series data.

        Returns
        -------
        :return data -- array | The raw time series data as 1D or 2D numpy array

        Examples
        --------
        data = extractor.get_raw_data('H', 'SITE_1')
        """

    @abstractmethod
    def get_stat_vector(self, site):
        """
        Query to extract an array that defines invalid time series data at a given location.

        Arguments
        ---------
        :param site -- string | Location at which to extract the time series data.

        Returns
        -------
        :return stat -- logical array | True if time series data is invalid (i.e cell is dry).

        Examples
        --------
        stat = extractor.get_stat_vector('SITE_1')
        """

    @abstractmethod
    def get_z_layer_faces(self, site):
        """
        Query to extract an array that defines the vertical layer faces of a 3D model at a given location.

        Arguments
        ---------
        :param site -- string | Location at which to extract the time series data.

        Returns
        -------
        :return lfz -- array | Vertical layer faces. If model is 2D returns None.

        Examples
        --------
        lfz = extractor.get_z_layer_faces('SITE_1')
        """

    @abstractmethod
    def get_integral_data(self, site, datum, limits):
        """
        Query to extract data for vertical integration at a given location. Principle data is the
        integral limit (z2 - z1) and dz for each 3D model cell at the location.

        Arguments
        ---------
        :param site -- string | Location at which to extract the time series data.
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

        Returns
        -------
        :return (z2_z1, dz) -- tuple | The elevation limits (z2 - z1) & dz for each 3D cell/node at a given location
        """

    @abstractmethod
    def get_data(self, variable, site, datum='sigma', limits=(0, 1)):
        """
        Query to extract time series data at a given location. If model data is 3D then it is
        depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param site -- string | Location at which to extract the time series data.

        Keyword Arguments
        -----------------
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

        Returns
        -------
        :return data -- array | Time series data at a given location.
        """

    @abstractmethod
    def __prepare_file_handle__(self):
        """Command which prepares the file handle for the extractor class"""

    @abstractmethod
    def __prepare_time_vector__(self):
        """Command which prepares the result time stamp vector relative to python epoch"""

    @abstractmethod
    def __prepare_locations__(self):
        """Command which prepares locations found in result time series file"""


class FvTimeSeries(TimeSeries):

    def __init__(self, file):
        super(FvTimeSeries, self).__init__(file)

    @Expression.decorator
    def get_raw_data(self, variable, site):
        return self.nc[site][variable][:].data.transpose()

    def get_stat_vector(self, site):
        return self.get_raw_data('stat', site) == 0

    def get_z_layer_faces(self, site):
        return self.get_raw_data('layerface_Z', site)

    def get_integral_data(self, site, datum, limits):
        # Get z layer faces
        lfz = self.get_z_layer_faces(site)

        # Get water level (wl) and bed level (bl) for each time step
        wli, bli = 0, lfz.shape[0]-1
        wl, bl = lfz[wli, :], lfz[bli, :]

        # Determine integral limits z1 and z2 for each time step using wl, bl and the limits
        if datum == 'sigma':
            z1 = limits[0] * (wl - bl) + bl
            z2 = limits[1] * (wl - bl) + bl
        elif datum == 'height':
            z1 = limits[0] + bl
            z2 = limits[1] + bl
        elif datum == 'depth':
            z1 = wl - limits[1]
            z2 = wl - limits[0]
        elif datum == 'elevation':
            z1 = limits[0]
            z2 = limits[1]
        else:
            return None

        # Create integral limits, filtering z2 and z1 above and below water level or bed level
        z1 = np.minimum(np.maximum(z1, bl), wl)
        z2 = np.minimum(np.maximum(z2, bl), wl)

        # Squeeze out middle value of each vertical layer face
        lfz = np.maximum(lfz, np.tile(z1, (lfz.shape[0], 1)))
        lfz = np.minimum(lfz, np.tile(z2, (lfz.shape[0], 1)))

        # Get upper z layer face and lower z layer face for each 3D cell
        ul = np.delete(lfz, bli, axis=0)
        ll = np.delete(lfz, wli, axis=0)
        dz = ul - ll

        # Clean up integral limit to avoid division by zero
        z2_z1 = z2 - z1
        mask = z2_z1 == 0

        # Return integral limit of each 2D cell and dz of each 3D cell contained within integral limit
        return np.ma.masked_array(data=z2_z1, mask=mask, fill_value=np.nan), dz

    @Expression.decorator
    def get_data(self, variable, site, datum='sigma', limits=(0, 1)):
        # Get the raw data
        data = self.get_raw_data(variable, site)
        stat = self.get_stat_vector(site)

        # If data is 2D depth average
        if data.ndim == 2:
            # Get the integral limits
            z2_z1, dz = self.get_integral_data(site, datum, limits)

            # Update stat vector with invalid limits
            stat = (stat | z2_z1.mask)

            # Integrate the data w.r.t z
            data = np.nansum(data*dz, axis=0) * (1 / z2_z1)

        # Return the data
        return np.ma.masked_array(data=data, mask=stat, fill_value=np.nan)

    def __prepare_file_handle__(self):
        self.nc = Dataset(self.file, 'r')

    def __prepare_time_vector__(self):
        # Prepare time vector relative to python epoch
        fv_time_vector = self.nc['ResTime'][:].data
        fv_epoch = dt.datetime(1990, 1, 1, 10).timestamp()
        fv_time_delta = fv_time_vector*3600

        self.time_vector = fv_time_delta + fv_epoch
        self.nt = self.time_vector.size

    def __prepare_locations__(self):
        self.locations = {}
        for group in self.nc.groups.keys():
            xp = self.nc[group]['x'][:].data[0]
            yp = self.nc[group]['y'][:].data[0]
            self.locations[group] = [xp, yp]

    def get_geometry(self, site):
        z = self.get_z_layer_faces(site)
        t = np.tile(self.time_vector, (z.shape[0], 1))

        tc = np.hstack((t[:, [0]], 0.5 * (t[:, :-1] + t[:, 1:]), t[:, [-1]]))
        zc = np.hstack((z[:, [0]], 0.5 * (z[:, :-1] + z[:, 1:]), z[:, [-1]]))

        node_x, node_y = tc.ravel('F'), zc.ravel('F')

        nz, nt = tc.shape
        ii = np.arange(nz - 1)
        jj = np.arange(nt - 1)

        jj, ii = np.meshgrid(jj, ii)
        tl = np.ravel(ii + nz * jj, 'F')
        tr = np.ravel(ii + nz * (jj + 1), 'F')

        cell_node = np.vstack((tl, tl + 1, tr + 1, tr))
        cell_node = cell_node.transpose()

        return Mesh(node_x, node_y, cell_node)

    # Inherit doc strings (needs to be done a better way)
    get_raw_data.__doc__ = TimeSeries.get_raw_data.__doc__
    get_stat_vector.__doc__ = TimeSeries.get_stat_vector.__doc__
    get_z_layer_faces.__doc__ = TimeSeries.get_z_layer_faces.__doc__
    get_integral_data.__doc__ = TimeSeries.get_integral_data.__doc__

    get_data.__doc__ = TimeSeries.get_data.__doc__

    __prepare_file_handle__.__doc__ = TimeSeries.__prepare_file_handle__.__doc__
    __prepare_time_vector__.__doc__ = TimeSeries.__prepare_time_vector__.__doc__
    __prepare_locations__.__doc__ = TimeSeries.__prepare_locations__.__doc__



import numpy as np
import datetime as dt
import inspect

__version__ = "0.0.2"
__author__ = "toby.devlin@bmtglobal.com, jonah.chorley@bmtglobal.com"
__aus_date__ = "%d/%m/%Y %H:%M:%S"


def dump(obj):
    """
    A function for printing the methods and properties (instance variables) of an object:

        dump(obj)

    :param obj: Any python object
    """
    methods = []
    properties = []

    for attr in dir(obj):
        if hasattr(obj, attr) and '__' not in attr:
            if callable(getattr(obj, attr)):
                methods.append(attr)
            else:
                properties.append(attr)

    print('Methods:')
    for attr in methods:
        print("obj.{}".format(attr))
    print('\nProperties:')
    for attr in properties:
        print("obj.{}".format(attr))


def get_main():
    frames = [stack[0] for stack in inspect.stack()]
    for frame in frames:
        module = inspect.getmodule(frame)
        if module.__name__ == '__main__':
            return module

    return None


def get_object(name):
    module = get_main()
    if hasattr(module, name):
        return getattr(module, name)
    else:
        return None


def get_date_time(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    # Basic input parameters
    in_array = np.array(in_array, ndmin=1)
    arr_type = in_array.dtype.type
    arr_shape = in_array.shape

    # Initialize conversion logic
    from_str = False
    from_num = False

    # Determine conversion logic
    if arr_type is np.str_:
        from_str = True
    else:
        from_num = True

    # Initialize output array
    out_array = np.empty(arr_shape, dtype=np.object)

    # If from string -> datetime
    if from_str:
        for ii in range(in_array.size):
            out_array[ii] = dt.datetime.strptime(in_array[ii], fmt)

    # If from number -> datetime
    if from_num:
        epoch = dt.datetime(1970, 1, 1)
        in_array = in_array.astype(np.float64)
        for ii in range(in_array.size):
            delta = dt.timedelta(seconds=in_array[ii])
            out_array[ii] = epoch + delta

    return out_array


def get_date_str(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    # Basic input parameters
    in_array = np.array(in_array, ndmin=1)
    arr_type = in_array.dtype.type
    arr_shape = in_array.shape

    # Convert any numbers to datetime
    if arr_type is not np.object_:
        in_array = get_date_time(in_array)

    # Initialize output array
    str_check = in_array[0].strftime(fmt)
    d_type = '<U{:d}'.format(len(str_check))
    out_array = np.empty(arr_shape, dtype=d_type)

    # Convert datetime to string
    for ii in range(in_array.size):
        out_array[ii] = in_array[ii].strftime(fmt)

    return out_array


def get_date_num(in_array, fmt='%d/%m/%Y %H:%M:%S'):
    # Basic input parameters
    in_array = np.array(in_array, ndmin=1)
    arr_type = in_array.dtype.type
    arr_shape = in_array.shape

    # Convert any numbers to datetime
    if arr_type is not np.object_:
        in_array = get_date_time(in_array, fmt)

    # Initialize output array
    out_array = np.empty(arr_shape, dtype=np.float64)

    # Convert datetime to number
    epoch = dt.datetime(1970, 1, 1)
    for ii in range(in_array.size):
        delta = in_array[ii] - epoch
        out_array[ii] = delta.total_seconds()

    return out_array


class Expression:

    @classmethod
    def decorator(cls, function_call):

        def wrapper(*args, **kwargs):
            # Basic preparation
            args = list(args)
            expression = None
            index = None

            # Find expressions in input arguments
            for ii, arg in enumerate(args):
                if type(arg).__name__ == cls.__name__:
                    expression = arg
                    index = ii

            # Evaluate output based on expression
            if expression is None:
                output = function_call(*args, **kwargs)
            else:
                eval_args = list()
                for ii in range(len(expression.variables)):
                    args[index] = expression.variables[ii]
                    if expression.is_defined[ii]:
                        eval_args.append(get_object(args[index]))
                    else:
                        eval_args.append(function_call(*args, **kwargs))
                output = expression.evaluate(*eval_args)

            return output

        return wrapper

    def __init__(self, expression):
        self.expression = expression
        self.__interpret__()

    def __interpret__(self):
        # Define legal operations, functions and characters
        ops = ['**', '*', '/', '+', '-']
        chars = ['[', ']', '(', ')', ' ', "'", ':']
        funcs = ['hyp', 'dot', 'vec', 'sum']
        calls = ['np.hypot', 'np.dot', 'np.column_stack', 'np.sum']

        # Determine variables that exist in expression string
        string = self.expression
        for op in ops:
            string = string.replace(op, ',')
        for other in funcs + chars:
            string = string.replace(other, '')
        self.variables = string.split(',')

        # Remove constants
        for string in self.variables:
            try:
                float(string)
                is_num = True
            except ValueError:
                is_num = False
            if is_num:
                self.variables.remove(string)

        # Find if any variables that are defined in main name space
        self.is_defined = list()
        for variable in self.variables:
            handle = get_object(variable)
            if handle is None:
                self.is_defined.append(False)
            else:
                self.is_defined.append(True)

        # Determine the evaluation string
        argument_fmt = 'args[{}]'
        self.string = self.expression
        for aa, variable in enumerate(self.variables):
            place_holder = variable
            replace_with = argument_fmt.format(aa)
            self.string = self.string.replace(place_holder, replace_with)
        for aa, func in enumerate(funcs):
            place_holder = func
            replace_with = calls[aa]
            self.string = self.string.replace(place_holder, replace_with)

    def evaluate(self, *args):
        return eval(self.string)


def unsupported_decorator(function_call):

    def wrapper(*args):
        name = function_call.__name__
        message = '{} is currently not supported'.format(name)
        print(message)

    return wrapper


def timer_decorator(function_call):

    name = function_call.__name__
    message = '{} took {}s to execute'

    def wrapper(*args, **kwargs):
        ts = dt.datetime.now().timestamp()
        output = function_call(*args, **kwargs)
        te = dt.datetime.now().timestamp()
        print(message.format(name, te - ts))

        return output

    return wrapper



"""A module defining all objects used to visualise model result data extracts"""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from abc import ABC, abstractmethod
from tfv.misc import get_date_time, Expression
from matplotlib.animation import FFMpegWriter
from matplotlib.colorbar import ColorbarBase
from matplotlib.text import Text
from matplotlib.lines import Line2D
from matplotlib.quiver import Quiver
from matplotlib.patches import Polygon
from matplotlib.tri import Triangulation, TriContourSet
from matplotlib.cbook import silent_list
from matplotlib._tri import TriContourGenerator
from matplotlib.collections import PolyCollection, TriMesh, PathCollection
# Non-essential imports
from matplotlib import cm
from matplotlib.dates import DateFormatter
from matplotlib.colors import Normalize, BoundaryNorm

__version__ = "0.0.2"
__author__ = "toby.devlin@bmtglobal.com, jonah.chorley@bmtglobal.com"
__aus_date__ = "%d/%m/%Y %H:%M:%S"

mpl.use('Qt5Agg')
plt.interactive(True)

__axes__ = list()


# ----------------------------------------------- Interactive Objects --------------------------------------------------
class Slider:

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, values):
        self._values = np.array(values)
        self._index = np.argmin(np.abs(self.values - self.value))
        self._value = self.values[self.index]

        self.update()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self.index = np.argmin(np.abs(self.values - value))

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index):
        in_range = 0 <= index and index <= self.values.size - 1
        is_new = index != self.index
        if in_range and is_new:
            self._index = index
            self._value = self._values[index]

            self.update()

    def __init__(self, figure, values, callbacks=[], init_value=None, inc=1):
        self.figure = figure
        self._init_patches()
        self._init_events()

        self.inc = inc
        self._playing = False
        self._dragging = False

        self._index = None
        if init_value is None:
            self._value = values[0]
        else:
            self._value = init_value
        self.callbacks = callbacks
        self.values = values

        # Provide figure update method, depends on backend
        back_end = mpl.get_backend()
        if back_end == 'TkAgg':  # TkAgg
            self.fig_draw = self.figure.canvas.draw
        elif back_end == 'Qt5Agg':  # PyQt5
            self.fig_draw = self.figure.canvas.update

    def _init_patches(self):
        self.bbox = [0.02, 0.02, 0.3, 0.06]
        self.axes = self.figure.add_axes(self.bbox)
        self.axes.xaxis.set_visible(False)
        self.axes.yaxis.set_visible(False)
        self.axes.set_navigate(False)

        __axes__.append(self.axes)

        def make_rectangle(w, h):
            return np.array([[0, 0], [0, h], [w, h], [w, 0]]) - [w/2, h/2]

        def make_triangle(w, h, sign):
            return np.array( [[0, 0], [sign*w, h / 2], [0, h]]) - [sign*w/2, h/2]

        xy = make_rectangle(0.90, 0.10) + [0.50, 0.7]
        self.bar = Polygon(xy, edgecolor='black', facecolor='black')
        self.axes.add_patch(self.bar)

        xy = make_rectangle(0.025, 0.40) + [0.05, 0.7]
        self.ball = Polygon(xy, edgecolor='black', facecolor='white')
        self.axes.add_patch(self.ball)

        xy = make_triangle(0.05, 0.40, -1) + [0.40, 0.25]
        self.bwd = Polygon(xy, edgecolor='black', facecolor='green')
        self.axes.add_patch(self.bwd)

        xy = make_rectangle(0.05, 0.40) + [0.50, 0.25]
        self.stp = Polygon(xy, edgecolor='black', facecolor='red')
        self.axes.add_patch(self.stp)

        xy = make_triangle(0.05, 0.40, 1) + [0.60, 0.25]
        self.fwd = Polygon(xy, edgecolor='black', facecolor='green')
        self.axes.add_patch(self.fwd)

    def _init_events(self):
        callback_map = dict(button_press_event=self._on_press, motion_notify_event=self._on_motion,
                            button_release_event=self._on_release, key_press_event=self._on_key_press)
        self._cid = [self.figure.canvas.mpl_connect(event, callback) for event, callback in callback_map.items()]

    def _on_press(self, event):
        if event.inaxes != self.axes:
            return

        in_ball, _ = self.ball.contains(event)
        in_bar, _ = self.bar.contains(event)
        in_fwd, _ = self.fwd.contains(event)
        in_stop, _ = self.stp.contains(event)
        in_bwd, _ = self.bwd.contains(event)

        if in_ball:
            self._dragging = True
        elif in_bar:
            self.value = self._pos2val(event.xdata)
        elif in_fwd:
            self.play_fwd()
        elif in_bwd:
            self.play_bwd()
        elif in_stop:
            self.stop()

    def _on_motion(self, event):
        in_axes = event.inaxes == self.axes
        if not self._dragging or not in_axes:
            return
        self.value = self._pos2val(event.xdata)
        # self.figure.canvas.draw()

    def _on_release(self, event):
        self._dragging = False
        # self.figure.canvas.draw()

    def _on_key_press(self, event):
        if event.key == 'left':
            plt.ioff()
            self.prev()
            self.quick_draw()
            plt.ion()
        elif event.key == 'right':
            plt.interactive(False)
            self.next()
            self.quick_draw()
            plt.interactive(True)

    def _scale(self):
        return (self.values[-1] - self.values[0])/0.9

    def _x0(self):
        return self.bar.xy[0, 0]

    def _val2pos(self, value):
        return (value - self.values[0])/self._scale()+self._x0()

    def _pos2val(self, x):
        return (x-self._x0())*self._scale() + self.values[0]

    def subscribe(self, callback):
        self.callbacks.append(callback)
        self.update()

    def unsubscribe(self, callback):
        self.callbacks.remove(callback)

    def update(self):
        x = self._val2pos(self.value)
        dx = x - np.mean(self.ball.xy[:,0])
        self.ball.set_xy(self.ball.xy + [dx, 0])

        for callback in self.callbacks:
            callback(self.value)

    def next(self):
        self.index += self.inc

    def prev(self):
        self.index -= self.inc

    def play_fwd(self):
        self._playing = True
        plt.interactive(False)
        while self._playing:
            self.next()
            self.quick_draw()

    def play_bwd(self):
        self._playing = True
        plt.interactive(False)
        while self._playing:
            self.prev()
            self.quick_draw()

    def stop(self):
        self._playing = False
        plt.interactive(True)

    def quick_draw(self):
        for axes in __axes__:
            # Redraw the only items in frame
            axes.redraw_in_frame()

            # Draw x grid-lines
            x_grid = axes.get_xgridlines()
            for line in x_grid:
                axes.draw_artist(line)

            # Draw y grid-lines
            y_grid = axes.get_ygridlines()
            for line in y_grid:
                axes.draw_artist(line)

        # self.figure.canvas.update()
        self.fig_draw()
        self.figure.canvas.flush_events()


class Cursor:
    pass


class Viewer(object):
    """
    Class to explore/view time varying model result data

    Keyword Arguments
    -----------------
    :param size -- tuple | Size of underlying figure as (width, height) in specified units
    :param units -- string | Figure units i.e. 'mm', 'cm' or 'm'
    :param time_fmt -- string | Time format of title based on python datetime string formatting

    Returns
    -------
    :return viewer -- object | Viewer instance

    Examples
    --------
    >> viewer = Viewer(size=(240, 200), units='mm', time_fmt='%d/%m/%Y %H:%M:%S')

    >> viewer = Viewer(size=(30, 15), units='cm', time_fmt='%H:%M:%S')
    """

    __register__ = list()

    @property
    def time_current(self):
        return self.slider_bar.value

    @time_current.setter
    def time_current(self, time_current):
        self.slider_bar.value = time_current

    @property
    def time_vector(self):
        return self.slider_bar.values

    @time_vector.setter
    def time_vector(self, time_vector):
        if np.all(self.time_vector == np.arange(100)):
            self.slider_bar.values = time_vector
        else:
            t1 = np.hstack((self.time_vector, time_vector)).min()
            t2 = np.hstack((self.time_vector, time_vector)).max()
            dt = np.min((np.mean(np.diff(self.time_vector)), np.mean(np.diff(time_vector))))
            self.slider_bar.values = np.arange(t1, t2, dt)

    def __init__(self, size=(240, 120), units='mm', time_fmt='%d/%m/%Y %H:%M:%S'):
        """Initializes Viewer object with figure size, figure units & title time format"""

        self.size = size
        self.units = units
        self.time_fmt = time_fmt

        self.__prep_figure__()
        self.__prep_slider__()
        self.__prep_title__()

        self.__register__.append(self)

    def __prep_figure__(self):
        """Prepares underlying figure"""

        sf_map = {'mm': 12 / 300, 'cm': 12 / 30, 'm': 12 / 0.3}

        sf = sf_map[self.units]
        w = self.size[0]*sf
        h = self.size[1]*sf

        self.figure = plt.figure(figsize=(w, h))

    def __prep_slider__(self):
        """Initializes the Slider object"""

        self.slider_bar = Slider(self.figure, np.arange(100), callbacks=[])

    def __prep_title__(self):
        rec = [0.30, 0.96, 0.40, 0.04]
        self.title_axes = self.figure.add_axes(rec)
        self.title_axes.xaxis.set_visible(False)
        self.title_axes.yaxis.set_visible(False)
        self.title_axes.set_navigate(False)

        for spine in self.title_axes.spines.values():
            spine.set_color('white')

        self.title_axes.set_xlim(0, 2)
        self.title_axes.set_ylim(0, 2)
        __axes__.append(self.title_axes)

        self.title = Text(1, 1, '', fontsize=12, horizontalalignment='center', verticalalignment='center')
        self.title_axes._add_text(self.title)

        if self.time_fmt is not None:
            self.slider_bar.subscribe(self.__set_title_time__)

    def __set_title_time__(self, time_stamp):
        date_time = get_date_time(time_stamp)
        string = date_time[0].strftime(self.time_fmt)
        self.title.set_text(string)

    def subscribe(self, callback):
        self.slider_bar.subscribe(callback)

    def unsubscribe(self, callback):
        self.slider_bar.unsubscribe(callback)

    def print(self, file_path, fmt='png', dpi=300):
        """
        Prints Viewer objects figure to image file.

        Arguments
        ---------
        :param file_path -- string | File path to output image file

        Keyword Arguments
        -----------------
        :param fmt -- string | File format i.e 'png', 'pdf', 'jpg'
        :param dpi -- integer | Resolution in dots per inch (dpi)

        Examples
        --------
        >> viewer.print('Model_Result_Image.png', fmt='png', dpi=300)
        """

        self.slider_bar.axes.set_visible(False)
        self.figure.savefig(file_path, format=fmt, dpi=dpi)
        self.slider_bar.axes.set_visible(True)

    def animate(self, file_path, ts, te, dt,  fps=10, dpi=150):
        """
        Writes animation file for specified time interval.

        Arguments
        ---------
        :param file_path -- string | File path to output animation
        :param ts -- float | Start time as python time stamp
        :param te -- float | End time as python time stamp
        :param dt -- float | Time increment at which to save frame

        Keyword Arguments
        -----------------
        :param fps -- integer | Frames per second
        :param dpi -- integer | Resolution in dots per inch (dpi)
        """

        # Hide slider bar
        self.slider_bar.axes.set_visible(False)

        # Loop through time
        nt = int((te - ts) / dt)+1
        writer = FFMpegWriter(fps=fps)
        with writer.saving(self.figure, file_path, dpi):
            for ii in range(nt):
                self.time_current = (ts + ii*dt)
                writer.grab_frame()

        # Show slider bar
        self.slider_bar.axes.set_visible(True)


class ColourBar(ColorbarBase):

    def __init__(self, patch, location='bottom', offset=0.075, thickness=0.010, label=''):
        # Get target axes handle
        if type(patch).__name__ == 'TriContourSet':
            target = patch.ax
        else:
            target = patch.axes

        if offset > 0:
            offset_1 = offset
            offset_2 = 0
        else:
            offset_1 = 0
            offset_2 = offset

        # Determine rectangles for target axes & colour bar axes
        rec = target.get_position().extents
        if location == 'bottom':
            rec1 = [rec[0], rec[1] + offset_1, rec[2] - rec[0], rec[3] - rec[1] - offset_1]
            rec2 = [rec[0], rec[1] + offset_2, rec[2] - rec[0], thickness]
        elif location == 'top':
            rec1 = [rec[0], rec[1], rec[2] - rec[0], rec[3] - rec[1] - offset_1]
            rec2 = [rec[0], rec[3] - offset_2, rec[2] - rec[0], thickness]
        elif location == 'left':
            rec1 = [rec[0] + offset_1, rec[1], rec[2] - rec[0] - offset_1, rec[3] - rec[1]]
            rec2 = [rec[0] + offset_2, rec[1], thickness, rec[3] - rec[1]]
        elif location == 'right':
            rec1 = [rec[0], rec[1], rec[2] - rec[0] - offset_1, rec[3] - rec[1]]
            rec2 = [rec[2] - offset_2, rec[1], thickness, rec[3] - rec[1]]

        target.set_position(rec1)
        axes = target.figure.add_axes(rec2)

        # Set orientation
        if location == 'bottom' or location == 'top':
            orientation = 'horizontal'
        elif location == 'left' or location == 'right':
            orientation = 'vertical'

        # Initialize ColorbarBase
        super(ColourBar, self).__init__(axes, patch.cmap, patch.norm, orientation=orientation)

        # Finish formatting
        if orientation == 'horizontal':
            axes.xaxis.set_ticks_position(location)
            axes.xaxis.set_label_position(location)
            axes.set_xlabel(label)
        elif orientation == 'vertical':
            axes.yaxis.set_ticks_position(location)
            axes.yaxis.set_label_position(location)
            axes.set_ylabel(label)


# --------------------------------------------------- Visual Objects ---------------------------------------------------
class Visual(ABC):
    """
    An abstract base class for an object which visualizes a variable. The main purpose of a Visual sub-class is to
    initialize and store a plotting library specific graphics object for a given extract type i.e SheetPatch.

    The Visual object works by subscribing its set_time_current method to the Slider object. When self.set_time_current
    is called by the Slider, the Visual object selects a valid time based on the Extractor objects time_vector and updates itself.
    """

    @property
    def time_vector(self):
        return self.extractor.time_vector

    def __init__(self, axes, extractor, expression, **kwargs):
        # Set target axes
        self.axes = axes
        __axes__.append(self.axes)

        # Set extractor
        self.extractor = extractor
        self._time_current = self.time_vector[0]
        self._time_index = 0

        # Set expression
        if type(expression) is str:
            self.expression = Expression(expression)
        else:
            self.expression = [Expression(exp) for exp in expression]

        # Connect with viewer
        viewer = kwargs.pop('viewer', None)
        if viewer is None:
            viewer = Viewer.__register__[0]
        self.viewer = viewer
        self.viewer.time_vector = self.time_vector
        self.viewer.subscribe(self.set_time_current)

        # Initialize the graphics object
        self.__prep_graphics_obj__(**kwargs)

        # Zoom to the graphics object
        self.zoom()

    def get_time_current(self):
        return self._time_current

    def set_time_current(self, time):
        ii = np.argmin(np.abs(self.time_vector - time))

        if ii != self._time_index:
            self._time_index = ii
            self._time_current = self.time_vector[ii]
            self.__dynamic_update__()

    @abstractmethod
    def __get_data__(self):
        """Abstract method which returns object data based on the expression"""

    @abstractmethod
    def __prep_graphics_obj__(self, **kwargs):
        """Abstract method to initialize the graphics object which will be used to visualize the result"""

    @abstractmethod
    def __static_update__(self):
        """Abstract method which updates static components of graphics object"""

    @abstractmethod
    def __dynamic_update__(self):
        """Abstract method which updates dynamic components graphics object"""

    @abstractmethod
    def zoom(self):
        """Abstract method which zooms axis to extent of graphics object"""


class SheetPatch(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as collection of patches

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- object | Extractor object
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | SheetPatch instance

    Examples
    --------
    >> xtr = FvExtractor('my_file_2D.nc')
    >> sheet_h = SheetPatch(axes, xtr, 'H')

    >> xtr = FvExtractor('my_file_3D.nc')
    >> sheet_v = SheetPatch(axes, xtr, 'hyp(V_x, V_y)', 'sigma', [0, 1])
    """

    def __init__(self, axes, extractor, expression, datum='sigma', limits=(0, 1), **kwargs):
        self.datum = datum
        self.limits = limits

        # Call initialize method of super class
        super(SheetPatch, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self._time_index, self.datum, self.limits)
        if self.shading is 'flat':
            return self.extractor.get_sheet_cell(*args)
        elif self.shading is 'interp':
            return self.extractor.get_sheet_node(*args)

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.shading = kwargs.pop('shading', 'flat')

        # Get handles on sheet geometry
        node_x = self.extractor.node_x
        node_y = self.extractor.node_y
        cell_node = self.extractor.cell_node
        tri_cell_node = self.extractor.tri_cell_node

        # Instantiate graphics object based on shading type
        data = self.__get_data__()
        if self.shading is 'flat':
            xy = np.dstack((node_x[cell_node], node_y[cell_node]))
            self.patch = PolyCollection(xy, array=data, **kwargs)
        elif self.shading is 'interp':
            self.tri = Triangulation(node_x, node_y, triangles=tri_cell_node)
            self.patch = TriMesh(self.tri, array=data, antialiased=True, **kwargs)

            # Mask invalid triangles
            mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
            self.tri.set_mask(mask)

        # Add the graphics object to axes
        self.axes.add_collection(self.patch)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        self.patch.set_array(data)

        # Mask invalid triangles
        if self.shading is 'interp':
            mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
            self.tri.set_mask(mask)

    def zoom(self):
        self.axes.set_xlim([np.min(self.extractor.node_x), np.max(self.extractor.node_x)])
        self.axes.set_ylim([np.min(self.extractor.node_y), np.max(self.extractor.node_y)])


class SheetContour(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as contours

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- object | Extractor object
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | SheetContour instance

    Examples
    --------
    >> xtr = FvExtractor('my_file_2D.nc')
    >> sheet_h = SheetContour(axes, xtr, 'H')

    >> xtr = FvExtractor('my_file_3D.nc')
    >> sheet_v = SheetContour(axes, xtr, 'hyp(V_x, V_y)', 'sigma', [0, 1])
    """

    def __init__(self, axes, extractor, expression, datum='sigma', limits=(0, 1), **kwargs):
        self.datum = datum
        self.limits = limits

        # Call initialize method of super class
        super(SheetContour, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self._time_index, self.datum, self.limits)
        return self.extractor.get_sheet_node(*args)

    def __prep_graphics_obj__(self, **kwargs):
        # Get handles on sheet geometry
        node_x = self.extractor.node_x
        node_y = self.extractor.node_y
        tri_cell_node = self.extractor.tri_cell_node

        # Get triangular mesh and initialize cpp triangulation
        tri = Triangulation(node_x, node_y, triangles=tri_cell_node)
        self.cpp_tri = tri.get_cpp_triangulation()

        # Get data
        data = self.__get_data__()

        # Mask bad triangles
        mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
        self.cpp_tri.set_mask(mask)

        # Set contour limits/levels
        zlim = [data.min(), data.max()]

        clim = kwargs.pop('clim', [zlim[0], zlim[1]])
        self.levels = kwargs.pop('levels', 50)

        if type(self.levels) is int:
            levels = np.linspace(clim[0], clim[1], self.levels)
        else:
            levels = self.levels

        self.cont = TriContourSet(self.axes, tri, data.data, levels, filled=True, extend='both', **kwargs)

        if clim is not None:
            self.cont.set_clim(clim)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        # Get data
        data = self.__get_data__()

        # Mask bad triangles
        mask = np.any(data.mask[self.extractor.tri_cell_node], axis=1)
        self.cpp_tri.set_mask(mask)

        # Set contour limits/levels
        zlim = [data.min(), data.max()]
        clim = self.cont.get_clim()

        if type(self.levels) is int:
            levels = np.linspace(clim[0], clim[1], self.levels)
        else:
            levels = self.levels

        # self._contour_args(self, args, kwargs)
        self.cont.zmin = zlim[0]
        self.cont.zmax = zlim[1]

        # self._contour_level_args
        self.cont.levels = levels

        # self.__init__
        self.cont._process_levels()

        # Update cpp generator
        self.cont.cppContourGenerator = TriContourGenerator(self.cpp_tri, data)

        # Update Paths
        lowers, uppers = self.cont._get_lowers_and_uppers()
        segs, kinds = self.cont._get_allsegs_and_allkinds()

        zorder = self.cont.collections[0].zorder
        for collection in self.cont.collections:
            self.cont.ax.collections.remove(collection)
        self.cont.collections = silent_list('PathCollection')

        for level, level_upper, segs, kinds in \
                zip(lowers, uppers, segs, kinds):
            paths = self.cont._make_paths(segs, kinds)

            col = \
                PathCollection(
                    paths,
                    antialiaseds=(self.cont.antialiased,),
                    edgecolors=None,
                    alpha=self.cont.alpha,
                    transform=self.cont.get_transform(),
                    zorder=zorder
                )

            self.cont.ax.add_collection(col, autolim=False)
            self.cont.collections.append(col)
        self.cont.changed()

    def zoom(self):
        self.axes.set_xlim([np.min(self.extractor.node_x), np.max(self.extractor.node_x)])
        self.axes.set_ylim([np.min(self.extractor.node_y), np.max(self.extractor.node_y)])


class SheetVector(Visual):
    """
    Class for dynamic visualization of model result sheet extracts as gridded vector field

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- object | Extractor object
    :param expression -- tuple | Tuple of string expressions that defines vector (vec_x, vec_y)

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param vectors -- object | SheetVector instance

    Examples
    --------
    >> xtr = FvExtractor('my_file_2D.nc')
    >> vectors = SheetContour(axes, xtr, ['V_x', 'V_y'])

    >> xtr = FvExtractor('my_file_3D.nc')
    >> vectors = SheetContour(axes, xtr, ['V_x', 'V_y'], 'depth', [0, 1])
    """

    def __init__(self, axes, extractor, expression, datum, limits, **kwargs):
        self.datum = datum
        self.limits = limits

        self.xg = None
        self.yg = None
        self.grid_index = None

        # Call initialize method of super class
        super(SheetVector, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self._time_index, self.xg, self.yg, self.datum, self.limits)
        u = self.extractor.get_sheet_grid(self.expression[0], *args, grid_index=self.grid_index)
        v = self.extractor.get_sheet_grid(self.expression[1], *args, grid_index=self.grid_index)

        return u, v

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.resolution = kwargs.pop('resolution', 40)

        # Instantiate graphics object with no data, data will be added on draw event automatically
        self.quiver = Quiver(self.axes, np.mean(self.extractor.node_x), np.mean(self.extractor.node_y), 0, 0, **kwargs)

        # Add the graphics object to axes
        self.axes.add_collection(self.quiver)

        # Connect the regrid method to the draw event of the figure object.
        self._x_cid = self.axes.callbacks.connect('xlim_changed', self.__static_update__)
        self._y_cid = self.axes.callbacks.connect('ylim_changed', self.__static_update__)

    def __static_update__(self, event=None):
        xlim = self.axes.get_xlim()
        ylim = self.axes.get_ylim()

        self.xg = np.linspace(xlim[0], xlim[1], self.resolution)
        self.yg = np.linspace(ylim[0], ylim[1], self.resolution)
        self.grid_index = self.extractor.get_grid_index(self.xg, self.yg)

        u, v = self.__get_data__()

        x, y = [arr.flatten() for arr in np.meshgrid(self.xg, self.yg)]
        xy = np.hstack((x[:, np.newaxis], y[:, np.newaxis]))

        self.quiver.set_offsets(xy)
        self.quiver.set_UVC(u, v)

    def __dynamic_update__(self):
        u, v = self.__get_data__()
        self.quiver.set_UVC(u, v)

    def zoom(self):
        self.axes.set_xlim([np.min(self.extractor.node_x), np.max(self.extractor.node_x)])
        self.axes.set_ylim([np.min(self.extractor.node_y), np.max(self.extractor.node_y)])


class CurtainPatch(Visual):
    """
    Class for dynamic visualization of model result curtain extracts as collection of patches.

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.

    Returns
    -------
    :param curtain -- object | CurtainPatch instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> curtain = SheetContour(axes, xtr, polyline, 'SAL')
    """

    def __init__(self, axes, extractor, expression, polyline, **kwargs):
        self.polyline = polyline
        self.x_data = extractor.get_intersection_data(polyline)
        self.index = extractor.get_curtain_cell_index(polyline)

        # Call initialize method of super class
        super(CurtainPatch, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.geo = self.extractor.get_curtain_cell_geo(self._time_index, self.polyline, self.x_data, self.index)

        args = (self.expression, self._time_index, self.polyline, self.x_data, self.index)
        return self.extractor.get_curtain_cell(*args)

    def __prep_graphics_obj__(self, **kwargs):
        data = self.__get_data__()
        node_x, node_y, cell_node = self.geo

        xy = np.dstack((node_x[cell_node], node_y[cell_node]))
        self.patch = PolyCollection(xy, array=data, **kwargs)

        # Add the graphics object to axes
        self.axes.add_collection(self.patch)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        node_x, node_y, cell_node = self.geo

        xy = np.dstack((node_x[cell_node], node_y[cell_node]))

        self.patch.set_verts(xy)
        self.patch.set_array(data)

    def zoom(self):
        self.axes.set_xlim([np.min(self.geo[0]), np.max(self.geo[0])])
        self.axes.set_ylim([np.min(self.geo[1]), np.max(self.geo[1])])


class CurtainVector(Visual):
    """
    Class for dynamic visualization of model result curtain extracts as gridded vector field

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.

    Returns
    -------
    :param curtain -- object | CurtainPatch instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> curtain = SheetContour(axes, xtr, polyline, ['hyp(V_x, V_y)', 'W'])
    """

    def __init__(self, axes, extractor, expression, polyline, **kwargs):
        self.polyline = polyline
        self.x_data = extractor.get_intersection_data(polyline)
        self.index = extractor.get_curtain_cell_index(polyline)

        self.xg = np.array([0, 1])
        self.yg = np.array([0, 1])

        # Call initialize method of super class
        super(CurtainVector, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.geo = self.extractor.get_curtain_cell_geo(self._time_index, self.polyline, self.x_data, self.index)

        args = (self._time_index, self.polyline, self.xg, self.yg, self.x_data, self.index)
        u = self.extractor.get_curtain_grid(self.expression[0], *args)
        v = self.extractor.get_curtain_grid(self.expression[1], *args)

        return u, v

    def __prep_graphics_obj__(self, **kwargs):
        # Pop key word arguments which are not used by the graphics object
        self.resolution = kwargs.pop('resolution', 40)

        # Update geometry
        self.__get_data__()

        # Instantiate graphics object with no data, data will be added on draw event automatically
        self.quiver = Quiver(self.axes, np.mean(self.geo[0]), np.mean(self.geo[1]), 0, 0, **kwargs)

        # Add the graphics object to axes
        self.axes.add_collection(self.quiver)

        # Connect the regrid method to the draw event of the figure object.
        self._cid = self.axes.figure.canvas.mpl_connect('draw_event', lambda event: self.__dynamic_update__())

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        xlim = self.axes.get_xlim()
        ylim = self.axes.get_ylim()

        self.xg = np.linspace(xlim[0], xlim[1], self.resolution)
        self.yg = np.linspace(ylim[0], ylim[1], self.resolution)

        u, v = self.__get_data__()

        x, y = [arr.flatten() for arr in np.meshgrid(self.xg, self.yg)]
        xy = np.hstack((x[:, np.newaxis], y[:, np.newaxis]))

        self.quiver.set_offsets(xy)
        self.quiver.set_UVC(u, v)

    def zoom(self):
        self.axes.set_xlim([np.min(self.geo[0]), np.max(self.geo[0])])
        self.axes.set_ylim([np.min(self.geo[1]), np.max(self.geo[1])])


class ProfileCell(Visual):
    """
    Class for dynamic visualization of model result profile extracts as vertical line

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param point -- tuple | Point (x, y) of profile location.

    Returns
    -------
    :param profile -- object | ProfileCell instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> profile = ProfileCell(axes, xtr, point, 'SAL')
    """

    def __init__(self, axes, extractor, expression, point, **kwargs):
        self.point = point

        super(ProfileCell, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        self.elevation = self.extractor.get_profile_cell_geo(self._time_index, self.point)

        args = (self.expression, self._time_index, self.point)
        return self.extractor.get_profile_cell(*args)

    def __prep_graphics_obj__(self, **kwargs):
        data = self.__get_data__()
        self.line = Line2D(data, self.elevation, **kwargs)

        self.axes.add_line(self.line)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        data = self.__get_data__()
        self.line.set_data(data, self.elevation)

    def zoom(self):
        data = self.__get_data__()
        if data.min() != 0 and data.max() != 0:
            self.axes.set_xlim([data.min()*0.95, data.max()*1.05])
        self.axes.set_ylim([self.elevation.min()*0.95, self.elevation.max()*1.05])


class SeriesGlider(Visual):
    """
    Class for dynamic visualization of model result time series extracts with current time marked

    Arguments
    ---------
    :param axes -- object | Axes object.
    :param extractor -- object | Extractor object.
    :param expression -- string | String expression that defines variable.
    :param location -- tuple | Point (x, y) of profile location.

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param profile -- object | ProfileCell instance.

    Examples
    --------
    >> xtr = FvExtractor('my_file_3D.nc')
    >> profile = ProfileCell(axes, xtr, point, 'SAL')
    """

    def __init__(self, axes, extractor, expression, location, datum='sigma', limits=(0, 1), **kwargs):
        self.location = location
        self.datum = datum
        self.limits = limits

        super(SeriesGlider, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self.location, self.datum, self.limits)
        return self.extractor.get_data(*args)

    def __prep_graphics_obj__(self, **kwargs):
        self.date_fmt = kwargs.pop('date_fmt', '%d/%m/%Y')

        x = self.extractor.time_vector
        x = get_date_time(x)
        y = self.__get_data__()

        self.time_series = Line2D(x, y, **kwargs)
        self.axes.add_line(self.time_series)

        glider_spec = dict(linewidth=0.8, linestyle='--', color='black')
        self.glider = Line2D([x[0], x[0]], [-10**20, 10**20], **glider_spec)
        self.axes.add_line(self.glider)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        tc = self.get_time_current()
        x = np.array([tc, tc])
        x = get_date_time(x)
        self.glider.set_xdata(x)

    def zoom(self):
        x, y = self.time_series.get_data()
        xlim = [x.min(), x.max()]
        ylim = [y.min(), y.max()]
        self.axes.set_xlim(xlim)
        self.axes.set_ylim(ylim)


class HovmollerGlider(Visual):

    def __init__(self, axes, extractor, expression, location, **kwargs):
        self.location = location

        super(HovmollerGlider, self).__init__(axes, extractor, expression, **kwargs)

    def __get_data__(self):
        args = (self.expression, self.location)
        return self.extractor.get_raw_data(*args).ravel('F')

    def __prep_graphics_obj__(self, **kwargs):
        self.mesh = self.extractor.get_geometry(self.location)
        self.mesh.node_x = get_date_time(self.mesh.node_x)
        self.mesh.node_x = mdates.date2num(self.mesh.node_x)

        node_x = self.mesh.node_x[self.mesh.cell_node]
        node_y = self.mesh.node_y[self.mesh.cell_node]
        verts = np.dstack((node_x, node_y))

        self.patch = PolyCollection(verts, array=self.__get_data__(), **kwargs)
        self.axes.add_collection(self.patch)

        glider_spec = dict(linewidth=0.8, linestyle='--', color='black')
        self.glider = Line2D(self.time_vector[[0, 0]], [-10 ** 20, 10 ** 20], **glider_spec)
        self.axes.add_line(self.glider)

    def __static_update__(self):
        pass

    def __dynamic_update__(self):
        tc = self.get_time_current()
        x = np.array([tc, tc])
        x = get_date_time(x)
        self.glider.set_xdata(x)

    def zoom(self):
        xlim = [self.mesh.node_x.min(), self.mesh.node_x.max()]
        ylim = [self.mesh.node_y.min(), self.mesh.node_y.max()]
        self.axes.set_xlim(xlim)
        self.axes.set_ylim(ylim)


# ------------------------------------------------ Delta Visual Objects ------------------------------------------------
class DeltaSheetPatch(SheetPatch):
    """
    Class for dynamic visualization of model result sheet extract differences as collection of patches

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | DeltaSheetPatch instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetPatch(axes, (base_xtr, dev_xtr), 'H')
    """

    def __init__(self, axes, extractors, expression, datum='sigma', limits=(0, 1), **kwargs):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetPatch, self).__init__(axes, self.extractors[0], expression, datum, limits, **kwargs)

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetPatch, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetPatch, self).__get_data__()

        return data_2 - data_1


class DeltaSheetContour(SheetContour):
    """
    Class for dynamic visualization of model result sheet extract differences as contours

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param sheet -- object | DeltaSheetContour instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetContour(axes, (base_xtr, dev_xtr), 'H')
    """

    def __init__(self, axes, extractors, expression, datum='sigma', limits=(0, 1), **kwargs):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetContour, self).__init__(axes, self.extractors[0], expression, datum, limits, **kwargs)

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetContour, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetContour, self).__get_data__()

        return data_2 - data_1


class DeltaSheetVector(SheetVector):
    """
    Class for dynamic visualization of model result sheet extract differences as gridded vector field

    Arguments
    ---------
    :param axes -- object | Axes object
    :param extractor -- tuple | Extractor objects (base, developed)
    :param expression -- string | Expression that defines variable

    Keyword Arguments
    -----------------
    :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
    :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

    Returns
    -------
    :param vector -- object | DeltaSheetVector instance

    Examples
    --------
    >> base_xtr = FvExtractor('my_file_base.nc')
    >> dev_xtr = FvExtractor('my_file_developed.nc')
    >> diff_sheet = DeltaSheetVector(axes, (base_xtr, dev_xtr), ['V_x', 'V_y'])
    """

    def __init__(self, axes, extractors, expression, datum='sigma', limits=(0, 1), **kwargs):
        # Store multiple extractor objects
        self.extractors = extractors

        # Initialize as super class
        super(DeltaSheetVector, self).__init__(axes, self.extractors[0], expression, datum, limits, **kwargs)

    def __get_data__(self):
        self.extractor = self.extractors[0]
        data_1 = super(DeltaSheetVector, self).__get_data__()

        self.extractor = self.extractors[1]
        data_2 = super(DeltaSheetVector, self).__get_data__()

        return data_2[0] - data_1[0], data_2[1] - data_1[1]

# --------------------------------------------- Static Plotting Functions ----------------------------------------------

def plot_mesh_data(mesh, data, stat, **kwargs):
    pass


import pkg_resources

# List of dependencies copied from requirements.txt
dependencies = \
    [
        'matplotlib>=3.1.1',
        'netCDF4>=1.5.1.2',
        'numpy>=1.17.0',
        'PyQt5>=5.13.0',
    ]

# Throw exception if correct dependencies are not met
pkg_resources.require(dependencies)



"""A module defining all Extractor classes. Extractors are the primary objects for model result data extraction"""

import os
import numpy as np
import datetime as dt
from abc import ABC, abstractmethod
from netCDF4 import Dataset
from tfv.geometry import Mesh
from tfv.misc import Expression, unsupported_decorator, get_date_num


__version__ = "0.0.2"
__author__ = "toby.devlin@bmtglobal.com, jonah.chorley@bmtglobal.com"
__aus_date__ = "%d/%m/%Y %H:%M:%S"


class Extractor(ABC):
    """
    A base class that defines the API for all model result subclasses. Examples of these subclasses
    might be model results such as a TUFLOW FV NetCDF file, a Hycom NetCDF file or a TUFLOW FV .dat file.
    """

    result_type = None

    def __init__(self, file, is_spherical):
        """Initializes Extractor object with a model results file i.e A TUFLOW FV netCDF4 results file."""

        # Store file path string as attribute
        self.file = file
        self.is_spherical = is_spherical

        # Prepare static Extractor attributes
        self.__prep_file_handle__()
        self.__prep_time_vector__()
        self.__prep_2d_geometry__()
        self.__prep_3d_geometry__()

    @abstractmethod
    def get_raw_data(self, variable, ii):
        """
        Query to extract raw data at a time step (if time-varying).

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | The time vector index at which to extract the data.

        Returns
        -------
        :return data -- array | The raw data as 1D or 2D numpy array

        Examples
        --------
        data = extractor.get_raw_data('H', 10)
        """
        pass

    @abstractmethod
    def get_stat_vector(self, ii):
        """
        Query to extract an array that defines invalid model data.

        Arguments
        ---------
        :param ii -- integer | Time index at which to extract the stat array.

        Returns
        -------
        :return stat -- logical array | True if model cells/nodes are invalid (i.e dry cells).

        Examples
        --------
        stat = extractor.get_stat_vector(10)
        """
        pass

    @abstractmethod
    def get_z_layer_faces(self, ii):
        """
        Query to extract an array that defines the vertical layer faces of a 3D model.

        Arguments
        ---------
        :param ii -- integer | Time index at which to extract the vertical layer faces.

        Returns
        -------
        :return lfz -- array | Vertical layer faces. If model is 2D returns None.

        Examples
        --------
        lfz = extractor.get_z_layer_faces(10)
        """
        pass

    @abstractmethod
    def get_integral_data(self, ii, datum, limits):
        """
        Query to extract data for vertical integration at given time step. Principle data is the
        integral limit (z2 - z1) for each 2D model cell/node and dz for each 3D model cell/node.

        Arguments
        ---------
        :param ii -- integer | Time index at which to extract the vertical layer faces.
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.

        Returns
        -------
        :return (z2_z1, dz) -- tuple | The elevation limits (z2 - z1) for each 2D cell/node & dz for each 3D cell/node
        """
        pass

    @abstractmethod
    def get_sheet_cell(self, variable, ii, datum='sigma', limits=(0, 1), z_data=None):
        """
        Query to extract data in a 2D map format ('sheet') at model cell centroids for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.

        Keyword Arguments
        -----------------
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.
        :param z_data -- tuple | Vertical integration data returned by self.get_integral_data

        Returns
        -------
        :return data -- 1D or 2D array | A 2D spatial 'sheet' of the relevant variable at time step ii.

        Examples
        --------
        >> data = extractor.get_sheet_cell('V_x', 10, 'depth', (0, 1))
        >> data = extractor.get_sheet_cell('V_x', 10, 'elevation', (-5, -4))

        >> z2_z1, dz = extractor.get_integral_data(15, 'sigma', (0, 1))
        >> data = extractor.get_sheet_cell('V_x', 15, z_data=(z2_z1, dz))
        """
        pass

    @abstractmethod
    def get_sheet_node(self, variable, ii, datum='sigma', limits=(0, 1), z_data=None):
        """
        Query to extract data in a 2D map format ('sheet') at model cell vertices for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.

        Keyword Arguments
        -----------------
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.
        :param z_data -- tuple | Vertical integration data returned by self.get_integral_data

        Returns
        -------
        :return data -- 1D or 2D array | A 2D spatial 'sheet' of the relevant variable at time step ii.

        Examples
        --------
        >> data = extractor.get_sheet_node('V_x', 10, 'depth', (0, 1))
        >> data = extractor.get_sheet_node('V_x', 10, 'elevation', (-5, -4))

        >> z2_z1, dz = extractor.get_integral_data(15, 'sigma', (0, 1))
        >> data = extractor.get_sheet_node('V_x', 15, z_data=(z2_z1, dz))
        """
        pass

    @abstractmethod
    def get_sheet_grid(self, variable, ii, grid_x, grid_y, datum='sigma', limits=(0, 1), z_data=None, grid_index=None):
        """
        Query to extract data in a 2D map format ('sheet') at fixed grid points for a given time step. If model
        data is 3D then it is depth-averaged according to the depth-averaging vertical datum and vertical limits.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.
        :param grid_x -- 1D array | Horizontal grid point values
        :param grid_y -- 1D array | Vertical grid point values

        Keyword Arguments
        -----------------
        :param datum -- string | Vertical depth-averaging datum i.e sigma, depth, height, elevation, top, bottom.
        :param limits -- tuple | Vertical depth-averaging limits relative to vertical datum.
        :param z_data -- tuple | Vertical integration data returned by self.get_integral_data
        :param grid_index -- 2D array | Mesh cell index for each grid point returned by self.get_grid_index

        Returns
        -------
        :return data -- 2D array | A gridded 2D spatial 'sheet' of the relevant variable at time step ii.

        Examples
        --------
        >> xlim = [extractor.node_x.min(), extractor.node_x.max()]
        >> ylim = [extractor.node_y.min(), extractor.node_y.max()]
        >> grid_x = np.linspace(xlim[0], xlim[1], 100)
        >> grid_y = np.linspace(ylim[0], ylim[1], 100)
        >> data = extractor.get_sheet_grid('H', 10, grid_x, grid_y)

        >> index = extractor.get_grid_index(grid_x, grid_y)
        >> data = extractor.get_sheet_grid('H', 10, grid_x, grid_y, grid_index=index)
        """
        pass

    @abstractmethod
    def get_curtain_cell(self, variable, ii, polyline, x_data=None, index=None):
        """
        Query to extract data in a 2D slice format ('curtain') at the cell centroids for a given time step. It does
        this along the polyline specified.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.
        :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.

        Keyword Arguments
        -----------------
        :param x_data -- tuple | Model edge intersection(x) data returned by self.get_intersections.
        :param index -- tuple | Curtain index data returned by self.get_curtain_cell_index.

        Returns
        -------
        :return data -- array | A 2D slice 'curtain' of the relevant variable at time step ii.

        Examples
        --------
        >> polyline = np.array([[0, 0], [10, 10]])
        >> data = extractor.get_curtain_cell('SAL', 10, polyline)

        >> x_data = extractor.get_intersections(polyline)
        >> index = extractor.get_curtain_cell_index(polyline, x_data=x_data)
        >> data = extractor.get_curtain_cell('V_x', 15, polyline, x_data=intersections, index=index)
        """
        pass

    @abstractmethod
    def get_curtain_edge(self, variable, ii, polyline, x_data=None, index=None):
        """
        Query to extract data in a 2D slice format ('curtain') at the cell centroids for a given time step. It does
        this along the polyline specified.

        This function is currently not supported.
        """
        pass

    @abstractmethod
    def get_curtain_grid(self, variable, ii, polyline, grid_x, grid_y, x_data=None, index=None, grid_index=None):
        """
        Query to extract data in a 2D slice format ('curtain') at fixed grid points for a given time step. It does
        this along the polyline specified.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.
        :param polyline -- 2D array | Polyline as [x, y] used to slice 3D data.
        :param grid_x -- 1D array | Horizontal grid point values
        :param grid_y -- 1D array | Vertical grid point values

        Keyword Arguments
        -----------------
        :param x_data -- tuple | Model edge intersection(x) data returned by self.get_intersections.
        :param index -- tuple | Curtain index data returned by extractor.get_curtain_cell_index.
        :param grid_index -- 2D array | Curtain cell index for each grid point

        Returns
        -------
        :return data -- array | A gridded 2D slice 'curtain' of the relevant variable at time step ii.

        Examples
        --------
        >> polyline = np.array([[0, 0], [10, 10]])
        >> grid_x = np.linspace(0, 100, 50)
        >> grid_y = np.linspace(-20, 0, 50)
        >> data = extractor.get_curtain_grid('SAL', 10, polyline, grid_x, grid_y)
        """
        pass

    @abstractmethod
    def get_profile_cell(self, variable, ii, point, index=None):
        """
        Query to extract data as 1D vertical profile of cells at the given time step.
        It does this at the point specified.

        Arguments
        ---------
        :param variable -- string | Name of time varying data set to be extracted.
        :param ii -- integer | Time index at which to extract the data.
        :param point -- tuple | Point (x, y) of profile location.

        Keyword Arguments
        -----------------
        :param index -- integer | Index of cell which contains the point.

        Returns
        -------
        :return data -- array | A 1D section of the vertical values of the relevant variable.

        Examples
        --------
        data = extractor.get_profile_cell('SAL', 10, (20, 30))
        """
        pass

    @abstractmethod
    def __prep_file_handle__(self):
        """Command which prepares the file handle for the extractor class"""

    @abstractmethod
    def __prep_time_vector__(self):
        """Command which prepares the result time stamp vector relative to python epoch"""

    @abstractmethod
    def __prep_2d_geometry__(self):
        """Command which prepares the result 2D mesh geometry"""

    @abstractmethod
    def __prep_3d_geometry__(self):
        """A command which prepares the result 3D mesh geometry"""


class FvExtractor(Extractor, Mesh):
    """
    Class that extracts data from a TUFLOW FV netCDF4 result file.

    Arguments
    ----------
    :param file -- string | Model result file path.

    Keyword Arguments
    -----------------
    :param is_spherical -- bool | True if model geometry defined in spherical coordinate reference system.

    Returns
    -------
    :return extractor -- object | FvExtractor instance
    """

    result_type = 'Cell-centred TUFLOWFV output'

    def __init__(self, file, is_spherical=True):
        super(FvExtractor, self).__init__(file, is_spherical)

    def get_raw_data(self, variable, ii):
        return self.nc[variable][ii, :].data

    def get_stat_vector(self, ii):
        return self.nc['stat'][ii, :].data == 0

    def get_z_layer_faces(self, ii):
        return self.nc['layerface_Z'][ii, :].data

    def get_integral_data(self, ii, datum, limits):
        lfz = self.get_z_layer_faces(ii)

        # Get water level (wl) and bed level (bl) of each 2D cell
        wl = lfz[self.wli]
        bl = lfz[self.bli]

        # Determine integral limits z1 and z2 for each 2D cell using wl, bl and the limits
        if datum == 'sigma':
            z1 = limits[0] * (wl - bl) + bl
            z2 = limits[1] * (wl - bl) + bl
        elif datum == 'height':
            z1 = limits[0] + bl
            z2 = limits[1] + bl
        elif datum == 'depth':
            z1 = wl - limits[1]
            z2 = wl - limits[0]
        elif datum == 'elevation':
            z1 = limits[0]
            z2 = limits[1]
        else:
            return None

        # Create integral limits, filtering z2 and z1 above and below water level or bed level
        z1 = np.minimum(np.maximum(z1, bl), wl)
        z2 = np.minimum(np.maximum(z2, bl), wl)

        # Squeeze out middle value of each vertical layer face
        lfz = np.maximum(lfz, z1[self.idx4])
        lfz = np.minimum(lfz, z2[self.idx4])

        # Get upper z layer face and lower z layer face for each 3D cell
        ul = np.delete(lfz, self.bli)
        ll = np.delete(lfz, self.wli)
        dz = ul - ll

        # Clean up integral limit (z2 - z1) to avoid division by zero
        z2_z1 = z2 - z1
        mask = z2_z1 == 0

        # Return integral limit of each 2D cell and dz of each 3D cell contained within integral limit
        return np.ma.masked_array(data=z2_z1, mask=mask, fill_value=np.nan), dz

    @Expression.decorator
    def get_sheet_cell(self, variable, ii, datum='sigma', limits=(0, 1), z_data=None):
        # Get the raw data
        data = self.get_raw_data(variable, ii)
        stat = self.get_stat_vector(ii)

        # Check if data is 3D
        if data.size == self.nc3:
            # Get integral data for depth averaging
            if z_data is None:
                z_data = self.get_integral_data(ii, datum, limits)
            z2_z1, dz = z_data

            # Update stat vector with invalid limits
            stat = (stat | z2_z1.mask)

            # Integrate the data w.r.t z
            data = np.bincount(self.idx2, data * dz) * (1 / z2_z1)

        # Reshape stat vector for 2D sheets i.e bed mass
        if data.shape != stat.shape:
            n = data.shape[1]
            stat = np.tile(stat, (n, 1))
            stat = np.transpose(stat)

        # Return the data
        return np.ma.MaskedArray(data=data, mask=stat, fill_value=np.nan)

    @Expression.decorator
    def get_sheet_node(self, variable, ii, datum='sigma', limits=(0, 1), z_data=None):
        # Get the raw data
        data = self.get_raw_data(variable, ii)
        stat = self.get_stat_vector(ii)

        # Check if data is 3D
        if data.size == self.nc3:
            # Get integral data for depth averaging
            if z_data is None:
                z_data = self.get_integral_data(ii, datum, limits)
            z2_z1, dz = z_data

            # Update stat vector for invalid limits (z2 - z1)
            stat = (stat | z2_z1.mask)

            # Integrate the data w.r.t z
            data = np.bincount(self.idx2, data * dz) * (1 / z2_z1)

        # Create copy of 2D node recovery weights
        weights = np.copy(self.weights)

        # Set weightings of invalid cells to zero
        weights[stat, :] = 0

        # Rescale weightings to account for discounted cells
        weights_sum = np.bincount(self.cell_node.flatten(), weights.flatten())
        stat = weights_sum == 0
        weights_sum[stat] = -999
        weights = weights / weights_sum[self.cell_node]

        if data.ndim == 1:
            # For each cell, calculate the weighted nodal data values
            tmp = np.tile(data, (4, 1)).transpose() * weights

            # Sum weighted data values for each node to get final vertex data
            data_node = np.bincount(self.cell_node.flatten(), tmp.flatten())
        else:
            data_node = np.empty((self.nv2, data.shape[1]), dtype=np.float64)
            for jj in range(data.shape[1]):
                # For each cell, calculate the weighted nodal data values
                tmp = np.tile(data[:, jj], (4, 1)).transpose() * weights

                # Sum weighted data values for each node to get final vertex data
                data_node[:, jj] = np.bincount(self.cell_node.flatten(), tmp.flatten())

            stat = np.tile(stat, (data.shape[1], 1))
            stat = np.transpose(stat)

        # Return the data
        return np.ma.MaskedArray(data=data_node, mask=stat, fill_value=np.nan)

    def get_sheet_grid(self, variable, ii, xg, yg, datum='sigma', limits=(0, 1), z_data=None, grid_index=None):
        # Get grid index
        if grid_index is None:
            grid_index = self.get_grid_index(xg, yg)
        mask = np.equal(grid_index, -999)
        valid = np.equal(mask, False)

        # Index mesh data using grid index
        grid_data = np.ma.MaskedArray(data=np.zeros(grid_index.shape) * np.nan, fill_value=np.nan, mask=mask)
        grid_data[valid] = self.get_sheet_cell(variable, ii, datum, limits, z_data)[grid_index[valid]]

        # Return gridded data
        return grid_data

    @Expression.decorator
    def get_curtain_cell(self, variable, ii, polyline, x_data=None, index=None):
        # Get prerequisite data
        stat = self.get_stat_vector(ii)
        data = self.get_raw_data(variable, ii)

        # Check if data is 3D
        assert data.size == self.nc3, "Data is not 3D"

        # Get edge intersection data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        _, _, idx = x_data

        # Get curtain index data
        if index is None:
            index = self.get_curtain_cell_index(polyline, x_data)
        line_index, cell_index = index

        # Return curtain data
        return np.ma.MaskedArray(data=data[cell_index], mask=stat[idx[line_index]], fill_value=np.nan)

    @unsupported_decorator
    def get_curtain_edge(self, variable, ii, polyline, x_data=None, index=None):
        pass

    def get_curtain_grid(self, variable, ii, polyline, xg, yg, x_data=None, index=None, grid_index=None):
        # Get grid index
        if grid_index is None:
            geo = self.get_curtain_cell_geo(ii, polyline, x_data)
            grid_index = Mesh(*geo).get_grid_index(xg, yg)
        mask = np.equal(grid_index, -999)
        valid = np.equal(mask, False)

        # Index mesh data using grid index
        grid_data = np.ma.MaskedArray(data=np.zeros(grid_index.shape) * np.nan, fill_value=np.nan, mask=mask)
        grid_data[valid] = self.get_curtain_cell(variable, ii, polyline, x_data, index)[grid_index[valid]]

        # Return gridded data
        return grid_data

    @Expression.decorator
    def get_profile_cell(self, variable, time, point, index=None):
        # Get the raw data
        data = self.get_raw_data(variable, time)

        # Get 2D cell index
        if index is None:
            index = self.get_cell_index(point[0], point[1])

        # Index the data
        data = data[self.idx2 == index]

        # Repeat to get discrete elements
        insert = np.arange(0, data.size)
        data = np.insert(data, insert, data[insert])

        return data

    def __prep_file_handle__(self):
        # Store netCDF4 Dataset handle
        self.nc = Dataset(self.file, 'r')

        # Assert the file type is correct
        self.result_type = getattr(self.nc, 'Type', None)
        assert self.result_type == type(self).result_type, "Invalid file"

    def __prep_time_vector__(self):
        # Define fv epoch relative to python epoch
        fv_epoch = dt.datetime(1990, 1, 1)
        fv_epoch = get_date_num(fv_epoch)

        # Prepare time vector relative to python epoch
        fv_time_vector = self.nc['ResTime'][:].data
        py_time_vector = fv_time_vector*3600 + fv_epoch

        # Store data as attributes
        self.time_vector = py_time_vector
        self.nt = self.time_vector.size

    def __prep_2d_geometry__(self):
        # Get basic data from file
        node_x = self.nc['node_X'][:].data
        node_y = self.nc['node_Y'][:].data
        cell_node = self.nc['cell_node'][:].data - 1

        # Identify null node indices
        cell_node[cell_node == -1] = -999

        # Initialize 2d geometry as Mesh
        Mesh.__init__(self, node_x, node_y, cell_node)

    def __prep_3d_geometry__(self):
        # Get basic data from file
        self.nz = self.nc['NL'][:].data

        # Prepare variables to define 3D mesh
        index = np.arange(self.nz.size)
        self.idx2 = np.repeat(index, self.nz)
        self.idx3 = np.cumsum(self.nz) - self.nz
        self.idx4 = np.repeat(index, self.nz + 1)
        self.wli = self.idx3 + index
        self.bli = self.wli + self.nz
        self.nc3 = self.idx2.size

    def get_curtain_cell_index(self, polyline, x_data=None):
        # Get cell x_data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        _, _, idx = x_data

        # Determine which cells in the polyline are valid
        is_valid = idx != -999

        # Prepare indexing variables for curtain
        idx = idx[is_valid]  # 2D cell index for each column in curtain
        nz = self.nz[idx]  # Number of 3D cells for each column in curtain
        n = int(np.sum(nz))  # Total number of 3D cells in curtain
        idx3 = np.cumsum(nz) - nz  # Index of top 3D cell for each column in curtain

        # Get 1D polyline index of the curtain cells (maps curtain cell to polyline segments)
        line_index = np.repeat(np.where(is_valid)[0], nz)

        # Get 3D cell index of the curtain cells (maps curtain cell to 3D model cell)
        cell_index = np.repeat(self.idx3[idx], nz) + (np.arange(n) - np.repeat(idx3, nz))

        return line_index, cell_index

    def get_curtain_cell_geo(self, ii, polyline, x_data=None, index=None):
        # Get layer face z
        lfz = self.get_z_layer_faces(ii)

        # Get edge intersection(x) data
        if x_data is None:
            x_data = self.get_intersection_data(polyline)
        x, y, idx = x_data

        # Get curtain index
        if index is None:
            index = self.get_curtain_cell_index(polyline, x_data)
        line_index, cell_index = index

        # Convert x & y into (m) if spherical
        if self.is_spherical:
            r_e = 6.371*10**6
            cf = np.pi/180
            x = (r_e*np.cos(y*cf))*(x*cf)
            y = r_e*(y*cf)

        # Prepare curtain mesh (x, y)
        dx = np.diff(x)
        dy = np.diff(y)
        ds = np.hypot(dx, dy)
        s = np.cumsum(ds)
        s = np.hstack((0, s))

        s1 = s[line_index]
        s2 = s[line_index + 1]
        z1 = np.delete(lfz, self.wli)[cell_index]
        z2 = np.delete(lfz, self.bli)[cell_index]

        # Get nodes and node indices that define each cell
        n = cell_index.size
        node_x = np.vstack((s1, s2, s2, s1)).ravel('F')
        node_y = np.vstack((z1, z1, z2, z2)).ravel('F')
        cell_node = np.arange(n * 4, dtype=np.int32).reshape(n, 4)

        # Return geometry
        return node_x, node_y, cell_node

    def get_profile_cell_geo(self, ii, point, index=None):
        # Get layer face z
        lfz = self.get_z_layer_faces(ii)

        # Get 2D cell index
        if index is None:
            index = self.get_cell_index(point[0], point[1])

        # Index faces of cell
        lfz = lfz[self.idx4 == index]

        # Repeat to represent discrete elements
        insert = np.arange(1, lfz.size-1)
        lfz = np.insert(lfz, insert, lfz[insert])

        return lfz

    def write_time_series_file(self, out_file, locations):
        """
        Query that writes 2D & 3D point time series data from a TUFLOW FV netCDF4 results file.

        Arguments
        ---------
        :param out_file -- string | File path to write time series file.
        :param locations -- dictionary | Extraction points as dict(SITE_1=(x, y), SITE_2=(x, y)).
        """

        # Check if out path exists
        if os.path.isfile(out_file):
            user_input = input(out_file + ' exists - enter Y/N to proceed:\n')
            while True:
                if user_input.upper() == 'Y':
                    os.remove(out_file)
                    break
                elif user_input.upper() == 'N':
                    break
                else:
                    user_input = input('Invalid user input please enter Y/N:\n')

        # Get list of valid variables
        variables, exclude = [], ['ResTime']
        for var in self.nc.variables.keys():
            if var not in exclude and self.nc[var].shape[0] == self.nt:
                variables.append(var)

        # Check variable dimensions, required for pre-allocation & indexing
        num_dim = {}
        for var in variables:
            if self.nc.variables[var].shape[1] == self.nc2:
                num_dim[var] = 2
            elif self.nc.variables[var].shape[1] == self.nc3:
                num_dim[var] = 3

        # Get 2D cell index (ii) of each point
        idx = {name: self.get_cell_index(point[0], point[1])[0] for name, point in locations.items()}

        # Preallocate memory for each location
        profile, size = {}, ()
        for name in locations:
            profile[name] = {}
            for var in variables:
                if var == 'layerface_Z':
                    size = (self.nt, self.nz[idx[name]]+1)
                elif num_dim[var] == 3:
                    size = (self.nt, self.nz[idx[name]])
                elif num_dim[var] == 2:
                    size = (self.nt, 1)
                data = np.empty(size, dtype=np.float32)
                mask = np.zeros(size, dtype=np.bool)
                profile[name][var] = np.ma.masked_array(data=data, mask=mask)

        # For each variable, for each time step, for each site get the data
        for var in variables:
            for ii in range(self.nt):
                data = self.get_raw_data(var, ii)
                stat = self.get_stat_vector(ii)
                for name in locations:
                    if var == 'layerface_Z':
                        profile[name][var][ii, :] = data[self.idx4 == idx[name]]
                    elif num_dim[var] == 3:
                        profile[name][var][ii, :] = data[self.idx2 == idx[name]]
                    elif num_dim[var] == 2:
                        profile[name][var][ii, :] = data[idx[name]]
                    profile[name][var].mask[ii, :] = stat[idx[name]]
                del data

        # Create netCDF4 outfile & add time dimension
        nc = Dataset(out_file, 'w', format='NETCDF4')
        nc.createDimension('time', self.nt)

        # Create time variable and write out data
        nc.createVariable('ResTime', 'f8', dimensions=('time', ), fill_value=np.nan)
        nc.variables['ResTime'][:] = self.nc.variables['ResTime'][:].data

        # For each location, for each variable, write out data
        dims = ()
        for name in locations:
            nc.createGroup(name)
            nc[name].createDimension('x', 1)
            nc[name].createDimension('y', 1)

            nc[name].createVariable('x', 'f8', dimensions=('x',))
            nc[name].createVariable('y', 'f8', dimensions=('x',))

            nc[name].variables['x'][:] = locations[name][0]
            nc[name].variables['y'][:] = locations[name][1]

            nc[name].createDimension('z', self.nz[idx[name]])
            nc[name].createDimension('lfz', self.nz[idx[name]]+1)

            for var in variables:
                if var == 'layerface_Z':
                    dims = ('time', 'lfz',)
                elif num_dim[var] == 3:
                    dims = ('time', 'z',)
                elif num_dim[var] == 2:
                    dims = ('time',)
                nc[name].createVariable(var, 'f8', dimensions=dims, fill_value=np.nan)
                nc[name].variables[var][:] = profile[name][var]

        nc.close()
        return

    def write_data_to_ascii(self, out_file, data, resolution, bbox=None, grid_index=None):
        """
        Query data at cell centroids to gridded raster ASCII file.

        Arguments
        ---------
        :param out_file -- string | File path of .asc file to write data to.
        :param data -- 1D array | Cell centred TUFLOW FV data as numpy array.
        :param resolution -- float | Grid resolution at which to output data.

        Keyword Arguments
        -----------------
        :param bbox -- list | The bounding box to trim the data with as [left, bottom, right, top].
        :param grid_index -- 2D array | Mesh cell index for each grid point returned by self.get_grid_index
        """

        # Calculate grid parameters
        if bbox is None:
            x_min, x_max = self.node_x.min(), self.node_x.max()
            y_min, y_max = self.node_y.min(), self.node_y.max()
        else:
            x_min, x_max = bbox[0], bbox[2]
            y_min, y_max = bbox[1], bbox[3]

        dx = x_max - x_min
        dy = y_max - y_min

        nc = int(np.ceil(dx / resolution))
        nr = int(np.ceil(dy / resolution))

        xg = np.linspace(x_min, x_max, nc)
        yg = np.linspace(y_min, y_max, nr)

        # Get grid index & flip (x increasing, y decreasing)
        if grid_index is None:
            grid_index = self.get_grid_index(xg, yg)
        grid_index = np.flipud(grid_index)
        valid = grid_index != -999

        # Index mesh data using grid index
        grid_data = np.ones(grid_index.shape) * np.nan
        grid_data[valid] = data[grid_index[valid]]
        grid_data[np.isnan(grid_data)] = -999

        # Specify header
        header = \
            [
                'ncols {:d}'.format(nc),
                'nrows {:d}'.format(nr),
                'xllcorner {:.7f}'.format(x_min-dx/2),
                'yllcorner {:.7f}'.format(y_min-dy/2),
                'cellsize {:.7f}'.format(resolution),
                'NODATA_value {:d}'.format(-999)
            ]

        # Prepare row format
        row_fmt = ['{' + 'a[{}]:.7f'.format(jj) + '}' for jj in range(nc)]
        row_fmt = ' '.join(row_fmt) + '\n'

        # Write ASCII file
        with open(out_file, 'w') as asc:
            asc.write('\n'.join(header)+'\n')
            for ii in range(nr):
                asc.write(row_fmt.format(a=grid_data[ii, :]))

    # Inherit doc strings (needs to be done a better way)
    get_raw_data.__doc__ = Extractor.get_raw_data.__doc__
    get_stat_vector.__doc__ = Extractor.get_stat_vector.__doc__
    get_z_layer_faces.__doc__ = Extractor.get_z_layer_faces.__doc__
    get_integral_data.__doc__ = Extractor.get_integral_data.__doc__

    get_sheet_cell.__doc__ = Extractor.get_sheet_cell.__doc__
    get_sheet_node.__doc__ = Extractor.get_sheet_node.__doc__
    get_sheet_grid.__doc__ = Extractor.get_sheet_grid.__doc__

    get_curtain_cell.__doc__ = Extractor.get_curtain_cell.__doc__
    get_curtain_edge.__doc__ = Extractor.get_curtain_edge.__doc__
    get_curtain_grid.__doc__ = Extractor.get_curtain_grid.__doc__
    get_profile_cell.__doc__ = Extractor.get_profile_cell.__doc__
